package org.apache.bcel.generic;

import java.util.Hashtable;
import java.util.Stack;
import java.util.Objects;
import java.util.Iterator;
import org.apache.bcel.classfile.Utility;
import org.apache.bcel.classfile.AnnotationEntry;
import java.util.Comparator;
import org.apache.bcel.classfile.LineNumber;
import org.apache.commons.lang3.ArrayUtils;
import org.apache.bcel.classfile.ParameterAnnotationEntry;
import org.apache.bcel.classfile.RuntimeVisibleParameterAnnotations;
import java.util.Arrays;
import org.apache.bcel.classfile.ParameterAnnotations;
import org.apache.bcel.classfile.LocalVariable;
import org.apache.bcel.classfile.CodeException;
import org.apache.bcel.classfile.Annotations;
import java.util.Collection;
import java.util.Collections;
import org.apache.bcel.classfile.ExceptionTable;
import org.apache.bcel.classfile.LocalVariableTable;
import org.apache.bcel.classfile.LineNumberTable;
import java.util.ArrayList;
import org.apache.bcel.classfile.Code;
import org.apache.bcel.classfile.Method;
import org.apache.bcel.classfile.Attribute;
import java.util.List;
import org.apache.bcel.classfile.LocalVariableTypeTable;
import org.apache.bcel.util.BCELComparator;

public class MethodGen extends FieldGenOrMethodGen
{
    private static BCELComparator bcelComparator;
    private String className;
    private Type[] argTypes;
    private String[] argNames;
    private int maxLocals;
    private int maxStack;
    private InstructionList il;
    private boolean stripAttributes;
    private LocalVariableTypeTable localVariableTypeTable;
    private final List<LocalVariableGen> variableList;
    private final List<LineNumberGen> lineNumberList;
    private final List<CodeExceptionGen> exceptionList;
    private final List<String> throwsList;
    private final List<Attribute> codeAttrsList;
    private List<AnnotationEntryGen>[] paramAnnotations;
    private boolean hasParameterAnnotations;
    private boolean haveUnpackedParameterAnnotations;
    private List<MethodObserver> observers;
    
    private static byte[] getByteCodes(final Method method) {
        final Code code = method.getCode();
        if (code == null) {
            throw new IllegalStateException(String.format("The method '%s' has no code.", method));
        }
        return code.getCode();
    }
    
    public static BCELComparator getComparator() {
        return MethodGen.bcelComparator;
    }
    
    public static int getMaxStack(final ConstantPoolGen cp, final InstructionList il, final CodeExceptionGen[] et) {
        final BranchStack branchTargets = new BranchStack();
        for (final CodeExceptionGen element : et) {
            final InstructionHandle handlerPc = element.getHandlerPC();
            if (handlerPc != null) {
                branchTargets.push(handlerPc, 1);
            }
        }
        int stackDepth = 0;
        int maxStackDepth = 0;
        InstructionHandle ih = il.getStart();
        while (ih != null) {
            final Instruction instruction = ih.getInstruction();
            final short opcode = instruction.getOpcode();
            final int delta = instruction.produceStack(cp) - instruction.consumeStack(cp);
            stackDepth += delta;
            if (stackDepth > maxStackDepth) {
                maxStackDepth = stackDepth;
            }
            if (instruction instanceof BranchInstruction) {
                final BranchInstruction branch = (BranchInstruction)instruction;
                if (instruction instanceof Select) {
                    final Select select = (Select)branch;
                    final InstructionHandle[] targets2;
                    final InstructionHandle[] targets = targets2 = select.getTargets();
                    for (final InstructionHandle target : targets2) {
                        branchTargets.push(target, stackDepth);
                    }
                    ih = null;
                }
                else if (!(branch instanceof IfInstruction)) {
                    if (opcode == 168 || opcode == 201) {
                        branchTargets.push(ih.getNext(), stackDepth - 1);
                    }
                    ih = null;
                }
                branchTargets.push(branch.getTarget(), stackDepth);
            }
            else if (opcode == 191 || opcode == 169 || (opcode >= 172 && opcode <= 177)) {
                ih = null;
            }
            if (ih != null) {
                ih = ih.getNext();
            }
            if (ih == null) {
                final BranchTarget bt = branchTargets.pop();
                if (bt == null) {
                    continue;
                }
                ih = bt.target;
                stackDepth = bt.stackDepth;
            }
        }
        return maxStackDepth;
    }
    
    public static void setComparator(final BCELComparator comparator) {
        MethodGen.bcelComparator = comparator;
    }
    
    public MethodGen(final int accessFlags, final Type returnType, final Type[] argTypes, String[] argNames, final String methodName, final String className, final InstructionList il, final ConstantPoolGen cp) {
        super(accessFlags);
        this.variableList = new ArrayList<LocalVariableGen>();
        this.lineNumberList = new ArrayList<LineNumberGen>();
        this.exceptionList = new ArrayList<CodeExceptionGen>();
        this.throwsList = new ArrayList<String>();
        this.codeAttrsList = new ArrayList<Attribute>();
        this.setType(returnType);
        this.setArgumentTypes(argTypes);
        this.setArgumentNames(argNames);
        this.setName(methodName);
        this.setClassName(className);
        this.setInstructionList(il);
        this.setConstantPool(cp);
        final boolean abstract_ = this.isAbstract() || this.isNative();
        InstructionHandle start = null;
        final InstructionHandle end = null;
        if (!abstract_) {
            start = il.getStart();
            if (!this.isStatic() && className != null) {
                this.addLocalVariable("this", ObjectType.getInstance(className), start, end);
            }
        }
        if (argTypes != null) {
            final int size = argTypes.length;
            for (final Type argType : argTypes) {
                if (Type.VOID == argType) {
                    throw new ClassGenException("'void' is an illegal argument type for a method");
                }
            }
            if (argNames != null) {
                if (size != argNames.length) {
                    throw new ClassGenException("Mismatch in argument array lengths: " + size + " vs. " + argNames.length);
                }
            }
            else {
                argNames = new String[size];
                for (int i = 0; i < size; ++i) {
                    argNames[i] = "arg" + i;
                }
                this.setArgumentNames(argNames);
            }
            if (!abstract_) {
                for (int i = 0; i < size; ++i) {
                    this.addLocalVariable(argNames[i], argTypes[i], start, end);
                }
            }
        }
    }
    
    public MethodGen(final Method method, final String className, final ConstantPoolGen cp) {
        this(method.getAccessFlags(), Type.getReturnType(method.getSignature()), Type.getArgumentTypes(method.getSignature()), null, method.getName(), className, ((method.getAccessFlags() & 0x500) == 0x0) ? new InstructionList(getByteCodes(method)) : null, cp);
        final Attribute[] attributes2;
        final Attribute[] attributes = attributes2 = method.getAttributes();
        for (Attribute a : attributes2) {
            final Attribute attribute = a;
            if (a instanceof Code) {
                final Code c = (Code)a;
                this.setMaxStack(c.getMaxStack());
                this.setMaxLocals(c.getMaxLocals());
                final CodeException[] ces = c.getExceptionTable();
                if (ces != null) {
                    for (final CodeException ce : ces) {
                        final int type = ce.getCatchType();
                        ObjectType cType = null;
                        if (type > 0) {
                            final String cen = method.getConstantPool().getConstantString(type, (byte)7);
                            cType = ObjectType.getInstance(cen);
                        }
                        final int endPc = ce.getEndPC();
                        final int length = getByteCodes(method).length;
                        InstructionHandle end;
                        if (length == endPc) {
                            end = this.il.getEnd();
                        }
                        else {
                            end = this.il.findHandle(endPc);
                            end = end.getPrev();
                        }
                        this.addExceptionHandler(this.il.findHandle(ce.getStartPC()), end, this.il.findHandle(ce.getHandlerPC()), cType);
                    }
                }
                final Attribute[] attributes3;
                final Attribute[] cAttributes = attributes3 = c.getAttributes();
                for (int length4 = attributes3.length, k = 0; k < length4; ++k) {
                    final Attribute cAttribute = a = attributes3[k];
                    if (a instanceof LineNumberTable) {
                        ((LineNumberTable)a).forEach(l -> {
                            final InstructionHandle ih = this.il.findHandle(l.getStartPC());
                            if (ih != null) {
                                this.addLineNumber(ih, l.getLineNumber());
                            }
                            return;
                        });
                    }
                    else if (a instanceof LocalVariableTable) {
                        this.updateLocalVariableTable((LocalVariableTable)a);
                    }
                    else if (a instanceof LocalVariableTypeTable) {
                        this.localVariableTypeTable = (LocalVariableTypeTable)a.copy(cp.getConstantPool());
                    }
                    else {
                        this.addCodeAttribute(a);
                    }
                }
            }
            else if (a instanceof ExceptionTable) {
                Collections.addAll(this.throwsList, ((ExceptionTable)a).getExceptionNames());
            }
            else if (a instanceof Annotations) {
                final Annotations runtimeAnnotations = (Annotations)a;
                runtimeAnnotations.forEach(element -> this.addAnnotationEntry(new AnnotationEntryGen(element, cp, false)));
            }
            else {
                this.addAttribute(a);
            }
        }
    }
    
    public void addAnnotationsAsAttribute(final ConstantPoolGen cp) {
        this.addAll(AnnotationEntryGen.getAnnotationAttributes(cp, super.getAnnotationEntries()));
    }
    
    public void addCodeAttribute(final Attribute a) {
        this.codeAttrsList.add(a);
    }
    
    public void addException(final String className) {
        this.throwsList.add(className);
    }
    
    public CodeExceptionGen addExceptionHandler(final InstructionHandle startPc, final InstructionHandle endPc, final InstructionHandle handlerPc, final ObjectType catchType) {
        if (startPc == null || endPc == null || handlerPc == null) {
            throw new ClassGenException("Exception handler target is null instruction");
        }
        final CodeExceptionGen c = new CodeExceptionGen(startPc, endPc, handlerPc, catchType);
        this.exceptionList.add(c);
        return c;
    }
    
    public LineNumberGen addLineNumber(final InstructionHandle ih, final int srcLine) {
        final LineNumberGen l = new LineNumberGen(ih, srcLine);
        this.lineNumberList.add(l);
        return l;
    }
    
    public LocalVariableGen addLocalVariable(final String name, final Type type, final InstructionHandle start, final InstructionHandle end) {
        return this.addLocalVariable(name, type, this.maxLocals, start, end);
    }
    
    public LocalVariableGen addLocalVariable(final String name, final Type type, final int slot, final InstructionHandle start, final InstructionHandle end) {
        return this.addLocalVariable(name, type, slot, start, end, slot);
    }
    
    public LocalVariableGen addLocalVariable(final String name, final Type type, final int slot, final InstructionHandle start, final InstructionHandle end, final int origIndex) {
        final byte t = type.getType();
        if (t != 16) {
            final int add = type.getSize();
            if (slot + add > this.maxLocals) {
                this.maxLocals = slot + add;
            }
            final LocalVariableGen l = new LocalVariableGen(slot, name, type, start, end, origIndex);
            final int i;
            if ((i = this.variableList.indexOf(l)) >= 0) {
                this.variableList.set(i, l);
            }
            else {
                this.variableList.add(l);
            }
            return l;
        }
        throw new IllegalArgumentException("Can not use " + type + " as type for local variable");
    }
    
    public void addObserver(final MethodObserver o) {
        if (this.observers == null) {
            this.observers = new ArrayList<MethodObserver>();
        }
        this.observers.add(o);
    }
    
    public void addParameterAnnotation(final int parameterIndex, final AnnotationEntryGen annotation) {
        this.ensureExistingParameterAnnotationsUnpacked();
        if (!this.hasParameterAnnotations) {
            final List<AnnotationEntryGen>[] parmList = new List[this.argTypes.length];
            this.paramAnnotations = parmList;
            this.hasParameterAnnotations = true;
        }
        final List<AnnotationEntryGen> existingAnnotations = this.paramAnnotations[parameterIndex];
        if (existingAnnotations != null) {
            existingAnnotations.add(annotation);
        }
        else {
            final List<AnnotationEntryGen> l = new ArrayList<AnnotationEntryGen>();
            l.add(annotation);
            this.paramAnnotations[parameterIndex] = l;
        }
    }
    
    public void addParameterAnnotationsAsAttribute(final ConstantPoolGen cp) {
        if (!this.hasParameterAnnotations) {
            return;
        }
        final Attribute[] attrs = AnnotationEntryGen.getParameterAnnotationAttributes(cp, this.paramAnnotations);
        if (attrs != null) {
            this.addAll(attrs);
        }
    }
    
    private Attribute[] addRuntimeAnnotationsAsAttribute(final ConstantPoolGen cp) {
        final Attribute[] attrs = AnnotationEntryGen.getAnnotationAttributes(cp, super.getAnnotationEntries());
        this.addAll(attrs);
        return attrs;
    }
    
    private Attribute[] addRuntimeParameterAnnotationsAsAttribute(final ConstantPoolGen cp) {
        if (!this.hasParameterAnnotations) {
            return Attribute.EMPTY_ARRAY;
        }
        final Attribute[] attrs = AnnotationEntryGen.getParameterAnnotationAttributes(cp, this.paramAnnotations);
        this.addAll(attrs);
        return attrs;
    }
    
    private void adjustLocalVariableTypeTable(final LocalVariableTable lvt) {
        final LocalVariable[] lv = lvt.getLocalVariableTable();
        for (final LocalVariable element : this.localVariableTypeTable.getLocalVariableTypeTable()) {
            for (final LocalVariable l : lv) {
                if (element.getName().equals(l.getName()) && element.getIndex() == l.getOrigIndex()) {
                    element.setLength(l.getLength());
                    element.setStartPC(l.getStartPC());
                    element.setIndex(l.getIndex());
                    break;
                }
            }
        }
    }
    
    public MethodGen copy(final String className, final ConstantPoolGen cp) {
        final Method m = ((MethodGen)this.clone()).getMethod();
        final MethodGen mg = new MethodGen(m, className, super.getConstantPool());
        if (super.getConstantPool() != cp) {
            mg.setConstantPool(cp);
            mg.getInstructionList().replaceConstantPool(super.getConstantPool(), cp);
        }
        return mg;
    }
    
    private void ensureExistingParameterAnnotationsUnpacked() {
        if (this.haveUnpackedParameterAnnotations) {
            return;
        }
        final Attribute[] attrs = this.getAttributes();
        ParameterAnnotations paramAnnVisAttr = null;
        ParameterAnnotations paramAnnInvisAttr = null;
        for (final Attribute attribute : attrs) {
            if (attribute instanceof ParameterAnnotations) {
                if (!this.hasParameterAnnotations) {
                    final List<AnnotationEntryGen>[] parmList = new List[this.argTypes.length];
                    Arrays.setAll(this.paramAnnotations = parmList, i -> new ArrayList());
                }
                this.hasParameterAnnotations = true;
                final ParameterAnnotations rpa = (ParameterAnnotations)attribute;
                if (rpa instanceof RuntimeVisibleParameterAnnotations) {
                    paramAnnVisAttr = rpa;
                }
                else {
                    paramAnnInvisAttr = rpa;
                }
                final ParameterAnnotationEntry[] parameterAnnotationEntries = rpa.getParameterAnnotationEntries();
                for (int j = 0; j < parameterAnnotationEntries.length; ++j) {
                    final ParameterAnnotationEntry immutableArray = rpa.getParameterAnnotationEntries()[j];
                    final List<AnnotationEntryGen> mutable = this.makeMutableVersion(immutableArray.getAnnotationEntries());
                    this.paramAnnotations[j].addAll(mutable);
                }
            }
        }
        if (paramAnnVisAttr != null) {
            this.removeAttribute(paramAnnVisAttr);
        }
        if (paramAnnInvisAttr != null) {
            this.removeAttribute(paramAnnInvisAttr);
        }
        this.haveUnpackedParameterAnnotations = true;
    }
    
    @Override
    public boolean equals(final Object obj) {
        return MethodGen.bcelComparator.equals(this, obj);
    }
    
    public List<AnnotationEntryGen> getAnnotationsOnParameter(final int i) {
        this.ensureExistingParameterAnnotationsUnpacked();
        if (!this.hasParameterAnnotations || i > this.argTypes.length) {
            return null;
        }
        return this.paramAnnotations[i];
    }
    
    public String getArgumentName(final int i) {
        return this.argNames[i];
    }
    
    public String[] getArgumentNames() {
        return this.argNames.clone();
    }
    
    public Type getArgumentType(final int i) {
        return this.argTypes[i];
    }
    
    public Type[] getArgumentTypes() {
        return this.argTypes.clone();
    }
    
    public String getClassName() {
        return this.className;
    }
    
    public Attribute[] getCodeAttributes() {
        return this.codeAttrsList.toArray(Attribute.EMPTY_ARRAY);
    }
    
    private CodeException[] getCodeExceptions() {
        final int size = this.exceptionList.size();
        final CodeException[] cExc = new CodeException[size];
        Arrays.setAll(cExc, i -> this.exceptionList.get(i).getCodeException(super.getConstantPool()));
        return cExc;
    }
    
    public CodeExceptionGen[] getExceptionHandlers() {
        return this.exceptionList.toArray(CodeExceptionGen.EMPTY_ARRAY);
    }
    
    public String[] getExceptions() {
        return this.throwsList.toArray(ArrayUtils.EMPTY_STRING_ARRAY);
    }
    
    private ExceptionTable getExceptionTable(final ConstantPoolGen cp) {
        final int size = this.throwsList.size();
        final int[] ex = new int[size];
        Arrays.setAll(ex, i -> cp.addClass(this.throwsList.get(i)));
        return new ExceptionTable(cp.addUtf8("Exceptions"), 2 + 2 * size, ex, cp.getConstantPool());
    }
    
    public InstructionList getInstructionList() {
        return this.il;
    }
    
    public LineNumberGen[] getLineNumbers() {
        return this.lineNumberList.toArray(LineNumberGen.EMPTY_ARRAY);
    }
    
    public LineNumberTable getLineNumberTable(final ConstantPoolGen cp) {
        final int size = this.lineNumberList.size();
        final LineNumber[] ln = new LineNumber[size];
        Arrays.setAll(ln, i -> this.lineNumberList.get(i).getLineNumber());
        return new LineNumberTable(cp.addUtf8("LineNumberTable"), 2 + ln.length * 4, ln, cp.getConstantPool());
    }
    
    public LocalVariableGen[] getLocalVariables() {
        final int size = this.variableList.size();
        final LocalVariableGen[] lg = new LocalVariableGen[size];
        this.variableList.toArray(lg);
        for (int i = 0; i < size; ++i) {
            if (lg[i].getStart() == null && this.il != null) {
                lg[i].setStart(this.il.getStart());
            }
            if (lg[i].getEnd() == null && this.il != null) {
                lg[i].setEnd(this.il.getEnd());
            }
        }
        if (size > 1) {
            Arrays.sort(lg, Comparator.comparingInt(LocalVariableGen::getIndex));
        }
        return lg;
    }
    
    public LocalVariableTable getLocalVariableTable(final ConstantPoolGen cp) {
        final LocalVariableGen[] lg = this.getLocalVariables();
        final int size = lg.length;
        final LocalVariable[] lv = new LocalVariable[size];
        Arrays.setAll(lv, i -> lg[i].getLocalVariable(cp));
        return new LocalVariableTable(cp.addUtf8("LocalVariableTable"), 2 + lv.length * 10, lv, cp.getConstantPool());
    }
    
    public LocalVariableTypeTable getLocalVariableTypeTable() {
        return this.localVariableTypeTable;
    }
    
    public int getMaxLocals() {
        return this.maxLocals;
    }
    
    public int getMaxStack() {
        return this.maxStack;
    }
    
    public Method getMethod() {
        final String signature = this.getSignature();
        final ConstantPoolGen cp = super.getConstantPool();
        final int nameIndex = cp.addUtf8(super.getName());
        final int signatureIndex = cp.addUtf8(signature);
        final byte[] byteCode = (byte[])((this.il != null) ? this.il.getByteCode() : null);
        LineNumberTable lnt = null;
        LocalVariableTable lvt = null;
        if (!this.variableList.isEmpty() && !this.stripAttributes) {
            this.updateLocalVariableTable(this.getLocalVariableTable(cp));
            this.addCodeAttribute(lvt = this.getLocalVariableTable(cp));
        }
        if (this.localVariableTypeTable != null) {
            if (lvt != null) {
                this.adjustLocalVariableTypeTable(lvt);
            }
            this.addCodeAttribute(this.localVariableTypeTable);
        }
        if (!this.lineNumberList.isEmpty() && !this.stripAttributes) {
            this.addCodeAttribute(lnt = this.getLineNumberTable(cp));
        }
        final Attribute[] codeAttrs = this.getCodeAttributes();
        int attrsLen = 0;
        for (final Attribute codeAttr : codeAttrs) {
            attrsLen += codeAttr.getLength() + 6;
        }
        final CodeException[] cExc = this.getCodeExceptions();
        final int excLen = cExc.length * 8;
        Code code = null;
        if (byteCode != null && !this.isAbstract() && !this.isNative()) {
            final Attribute[] attributes2;
            final Attribute[] attributes = attributes2 = this.getAttributes();
            for (final Attribute a : attributes2) {
                if (a instanceof Code) {
                    this.removeAttribute(a);
                }
            }
            code = new Code(cp.addUtf8("Code"), 8 + byteCode.length + 2 + excLen + 2 + attrsLen, this.maxStack, this.maxLocals, byteCode, cExc, codeAttrs, cp.getConstantPool());
            this.addAttribute(code);
        }
        final Attribute[] annotations = this.addRuntimeAnnotationsAsAttribute(cp);
        final Attribute[] parameterAnnotations = this.addRuntimeParameterAnnotationsAsAttribute(cp);
        ExceptionTable et = null;
        if (!this.throwsList.isEmpty()) {
            this.addAttribute(et = this.getExceptionTable(cp));
        }
        final Method m = new Method(super.getAccessFlags(), nameIndex, signatureIndex, this.getAttributes(), cp.getConstantPool());
        if (lvt != null) {
            this.removeCodeAttribute(lvt);
        }
        if (this.localVariableTypeTable != null) {
            this.removeCodeAttribute(this.localVariableTypeTable);
        }
        if (lnt != null) {
            this.removeCodeAttribute(lnt);
        }
        if (code != null) {
            this.removeAttribute(code);
        }
        if (et != null) {
            this.removeAttribute(et);
        }
        this.removeRuntimeAttributes(annotations);
        this.removeRuntimeAttributes(parameterAnnotations);
        return m;
    }
    
    public Type getReturnType() {
        return this.getType();
    }
    
    @Override
    public String getSignature() {
        return Type.getMethodSignature(super.getType(), this.argTypes);
    }
    
    @Override
    public int hashCode() {
        return MethodGen.bcelComparator.hashCode(this);
    }
    
    private List<AnnotationEntryGen> makeMutableVersion(final AnnotationEntry[] mutableArray) {
        final List<AnnotationEntryGen> result = new ArrayList<AnnotationEntryGen>();
        for (final AnnotationEntry element : mutableArray) {
            result.add(new AnnotationEntryGen(element, this.getConstantPool(), false));
        }
        return result;
    }
    
    public void removeCodeAttribute(final Attribute a) {
        this.codeAttrsList.remove(a);
    }
    
    public void removeCodeAttributes() {
        this.localVariableTypeTable = null;
        this.codeAttrsList.clear();
    }
    
    public void removeException(final String c) {
        this.throwsList.remove(c);
    }
    
    public void removeExceptionHandler(final CodeExceptionGen c) {
        this.exceptionList.remove(c);
    }
    
    public void removeExceptionHandlers() {
        this.exceptionList.clear();
    }
    
    public void removeExceptions() {
        this.throwsList.clear();
    }
    
    public void removeLineNumber(final LineNumberGen l) {
        this.lineNumberList.remove(l);
    }
    
    public void removeLineNumbers() {
        this.lineNumberList.clear();
    }
    
    public void removeLocalVariable(final LocalVariableGen l) {
        l.dispose();
        this.variableList.remove(l);
    }
    
    public void removeLocalVariables() {
        this.variableList.forEach(LocalVariableGen::dispose);
        this.variableList.clear();
    }
    
    public void removeLocalVariableTypeTable() {
        this.localVariableTypeTable = null;
    }
    
    public void removeNOPs() {
        if (this.il != null) {
            InstructionHandle next;
            for (InstructionHandle ih = this.il.getStart(); ih != null; ih = next) {
                next = ih.getNext();
                if (next != null && ih.getInstruction() instanceof NOP) {
                    try {
                        this.il.delete(ih);
                    }
                    catch (final TargetLostException e) {
                        for (final InstructionHandle target : e.getTargets()) {
                            for (final InstructionTargeter targeter : target.getTargeters()) {
                                targeter.updateTarget(target, next);
                            }
                        }
                    }
                }
            }
        }
    }
    
    public void removeObserver(final MethodObserver o) {
        if (this.observers != null) {
            this.observers.remove(o);
        }
    }
    
    public void removeRuntimeAttributes(final Attribute[] attrs) {
        for (final Attribute attr : attrs) {
            this.removeAttribute(attr);
        }
    }
    
    public void setArgumentName(final int i, final String name) {
        this.argNames[i] = name;
    }
    
    public void setArgumentNames(final String[] argNames) {
        this.argNames = argNames;
    }
    
    public void setArgumentType(final int i, final Type type) {
        this.argTypes[i] = type;
    }
    
    public void setArgumentTypes(final Type[] argTypes) {
        this.argTypes = argTypes;
    }
    
    public void setClassName(final String className) {
        this.className = className;
    }
    
    public void setInstructionList(final InstructionList il) {
        this.il = il;
    }
    
    public void setMaxLocals() {
        if (this.il != null) {
            int max = this.isStatic() ? 0 : 1;
            if (this.argTypes != null) {
                for (final Type argType : this.argTypes) {
                    max += argType.getSize();
                }
            }
            for (InstructionHandle ih = this.il.getStart(); ih != null; ih = ih.getNext()) {
                final Instruction ins = ih.getInstruction();
                if (ins instanceof LocalVariableInstruction || ins instanceof RET || ins instanceof IINC) {
                    final int index = ((IndexedInstruction)ins).getIndex() + ((TypedInstruction)ins).getType(super.getConstantPool()).getSize();
                    if (index > max) {
                        max = index;
                    }
                }
            }
            this.maxLocals = max;
        }
        else {
            this.maxLocals = 0;
        }
    }
    
    public void setMaxLocals(final int m) {
        this.maxLocals = m;
    }
    
    public void setMaxStack() {
        if (this.il != null) {
            this.maxStack = getMaxStack(super.getConstantPool(), this.il, this.getExceptionHandlers());
        }
        else {
            this.maxStack = 0;
        }
    }
    
    public void setMaxStack(final int m) {
        this.maxStack = m;
    }
    
    public void setReturnType(final Type returnType) {
        this.setType(returnType);
    }
    
    public void stripAttributes(final boolean flag) {
        this.stripAttributes = flag;
    }
    
    @Override
    public final String toString() {
        final String access = Utility.accessToString(super.getAccessFlags());
        String signature = Type.getMethodSignature(super.getType(), this.argTypes);
        signature = Utility.methodSignatureToString(signature, super.getName(), access, true, this.getLocalVariableTable(super.getConstantPool()));
        final StringBuilder buf = new StringBuilder(signature);
        for (final Attribute a : this.getAttributes()) {
            if (!(a instanceof Code) && !(a instanceof ExceptionTable)) {
                buf.append(" [").append(a).append("]");
            }
        }
        if (!this.throwsList.isEmpty()) {
            for (final String throwsDescriptor : this.throwsList) {
                buf.append("\n\t\tthrows ").append(throwsDescriptor);
            }
        }
        return buf.toString();
    }
    
    public void update() {
        if (this.observers != null) {
            for (final MethodObserver observer : this.observers) {
                observer.notify(this);
            }
        }
    }
    
    private void updateLocalVariableTable(final LocalVariableTable a) {
        this.removeLocalVariables();
        for (final LocalVariable l : a.getLocalVariableTable()) {
            InstructionHandle start = this.il.findHandle(l.getStartPC());
            final InstructionHandle end = this.il.findHandle(l.getStartPC() + l.getLength());
            if (null == start) {
                start = this.il.getStart();
            }
            this.addLocalVariable(l.getName(), Type.getType(l.getSignature()), l.getIndex(), start, end, l.getOrigIndex());
        }
    }
    
    static {
        MethodGen.bcelComparator = new BCELComparator() {
            @Override
            public boolean equals(final Object o1, final Object o2) {
                final FieldGenOrMethodGen THIS = (FieldGenOrMethodGen)o1;
                final FieldGenOrMethodGen THAT = (FieldGenOrMethodGen)o2;
                return Objects.equals(THIS.getName(), THAT.getName()) && Objects.equals(THIS.getSignature(), THAT.getSignature());
            }
            
            @Override
            public int hashCode(final Object o) {
                final FieldGenOrMethodGen THIS = (FieldGenOrMethodGen)o;
                return THIS.getSignature().hashCode() ^ THIS.getName().hashCode();
            }
        };
    }
    
    static final class BranchStack
    {
        private final Stack<BranchTarget> branchTargets;
        private final Hashtable<InstructionHandle, BranchTarget> visitedTargets;
        
        BranchStack() {
            this.branchTargets = new Stack<BranchTarget>();
            this.visitedTargets = new Hashtable<InstructionHandle, BranchTarget>();
        }
        
        public BranchTarget pop() {
            if (!this.branchTargets.empty()) {
                return this.branchTargets.pop();
            }
            return null;
        }
        
        public void push(final InstructionHandle target, final int stackDepth) {
            if (this.visited(target)) {
                return;
            }
            this.branchTargets.push(this.visit(target, stackDepth));
        }
        
        private BranchTarget visit(final InstructionHandle target, final int stackDepth) {
            final BranchTarget bt = new BranchTarget(target, stackDepth);
            this.visitedTargets.put(target, bt);
            return bt;
        }
        
        private boolean visited(final InstructionHandle target) {
            return this.visitedTargets.get(target) != null;
        }
    }
    
    static final class BranchTarget
    {
        final InstructionHandle target;
        final int stackDepth;
        
        BranchTarget(final InstructionHandle target, final int stackDepth) {
            this.target = target;
            this.stackDepth = stackDepth;
        }
    }
}
